/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.common.artifact;

import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;

public class QueryStringPayload implements Payload {

    private Map mParameterMap = new LinkedHashMap();

    public QueryStringPayload() {
    }

    public QueryStringPayload(final String pQueryString) {
        parseQueryString(pQueryString, false);
    }

    public static QueryStringPayload createFromEncodedString(String queryString) {
        QueryStringPayload payload = new QueryStringPayload();
        payload.parseQueryString(queryString, true);
        return payload;
    }

    private void parseQueryString(final String pQueryString, boolean urlEncoded) {
        // the problem here is that we use an decoded querystring,
        // so '=' and '&' are allowed as values...
        // better where to have an url encoded string
        // don't use split here, to avoid wrong parsing
        // here we use a look ahead approach to find correct key/value pairs
        int currPairStart = 0;
        int currPairEnd = 0;
        int currPairSplit = 0;
        // Sample String: REQUEST=GetCapab=iliti&rest&SERVICE=WMS==
        // Shall be parsed to:
        // 1 : key= 'REQUEST' value= 'GetCapab=iliti&rest'
        // 2 : key= 'SERVICE' value= 'WMS=='
        while (currPairStart < pQueryString.length()) {
            String name = "";
            String value = "";
            currPairSplit = pQueryString.indexOf("=", currPairStart);
            currPairSplit = currPairSplit > -1 ? currPairSplit : pQueryString.length();
            name = pQueryString.substring(currPairStart, currPairSplit);
            currPairStart = currPairSplit + 1;
            if (currPairStart < pQueryString.length()) {
                currPairEnd = findEnd(pQueryString, currPairStart);
                currPairEnd = currPairEnd > -1 ? currPairEnd : pQueryString.length();
                value = pQueryString.substring(currPairStart, currPairEnd);
                if (urlEncoded) {
                    try {
                    	name = URLDecoder.decode(name, "utf-8");
                        value = URLDecoder.decode(value, "utf-8");
                    } catch (UnsupportedEncodingException e) {
                        throw new RuntimeException(e);
                    }
                }
                currPairStart = currPairEnd + 1;
            }
            if (name.length() > 0) {
                addParameter(name, value);
            }

        }
    }

    private int findEnd(String pQueryString, int currStart) {
        // tries to find index of last & of a value: 'GetCapab=iliti&rest&'
        // in REQUEST=GetCapab=iliti&rest&SERVICE=WMS==
        int ampersand = pQueryString.indexOf("&", currStart);
        if (ampersand > 0) {
            int nextEq = pQueryString.indexOf("=", ampersand + 1);
            if (nextEq < 0) {
                return -1;
            }
            int nextampersand = pQueryString.indexOf("&", ampersand + 1);
            while (nextampersand > -1 && nextEq > nextampersand) {
                ampersand = nextampersand;
                nextampersand = pQueryString.indexOf("&", ampersand + 1);
            }
        }
        return ampersand;
    }

    public Iterator iterator() {
        return mParameterMap.entrySet().iterator();
    }

    public int size() {
        return mParameterMap.size();
    }

    // TODO JavaDoc that describes case insensitivity
    public String getParameterValue(final String pKey) {
        // Clarify: what is returned if param=value1,value2,value3? Just value1 or all values?
        if (pKey != null && pKey.length() > 0) {
            Object lParamValue = mParameterMap.get(new Key(pKey));
            return (lParamValue == null) ? null : (String) lParamValue;
        }
        return null;
    }

    public String[] getParameterValues(final String pKey) {
        String lValues = getParameterValue(pKey);
        if (lValues == null || lValues.equals("")) {
            return new String[] {};
        }
        String[] lSplittedValues = lValues.split(",");
        return lSplittedValues;
    }

    public void replaceParameter(final String pKey, final String pValue) {
        if (pKey != null && pKey.length() > 0) {
            if (pValue == null) {
                mParameterMap.remove(new Key(pKey));
            } else {
                mParameterMap.put(new Key(pKey), pValue);
            }
        }
    }

    public void addParameter(final String pKey, final String pValue) {
        if (pKey != null && pKey.length() > 0) {
            mParameterMap.put(new Key(pKey), pValue);
        }
    }

    /**
     * Concatenates List elements separated by a certain String
     *
     * @param pList
     * @param pSeparator
     * @return
     */
    public static String listToString(final Collection pList, final String pSeparator) {
        if (pList.size() == 0)
            return "";
        StringBuffer lCSV = new StringBuffer();
        for (Iterator it = pList.iterator(); it.hasNext();) {
            String tValue = (String) it.next();
            lCSV.append(tValue).append(pSeparator);
        }
        if (lCSV.length() > 0) {
            lCSV.delete(lCSV.length() - pSeparator.length(), lCSV.length());
        }
        return lCSV.toString();
    }

    public InputStream getAsStream() {
        return new ByteArrayInputStream(toString().getBytes());
    }

    public String toString() {

        MapEntryRenderer nonEncodingRender = new MapEntryRenderer() {
            public String render(Map.Entry tEntry) {
                return tEntry.getKey() + "=" + tEntry.getValue() + "&";
            }
        };


        return render(nonEncodingRender);
    }

    public String toEncodedString() {
        MapEntryRenderer encodingRender = new MapEntryRenderer() {
            public String render(Map.Entry tEntry) {
                return URLEncoder.encode(((Key)tEntry.getKey()).toString()) + "=" + URLEncoder.encode((String) tEntry.getValue()) + "&";
            }
        };


        return render(encodingRender);
    }

    private String render(MapEntryRenderer nonEncodingRender) {
        Iterator itr = mParameterMap.entrySet().iterator();
        StringBuffer qstr = new StringBuffer(mParameterMap.size() * 20);

        while (itr.hasNext()) {
            Map.Entry tEntry = (Map.Entry) itr.next();
            qstr.append(nonEncodingRender.render(tEntry));
        }
        if (qstr.length() > 0) {
            qstr.deleteCharAt(qstr.length() - 1);
        }
        return qstr.toString();
    }





    private static interface MapEntryRenderer {
        String render(Map.Entry kvp);
    }

}
