/*******************************************************************************
 * Copyright (C) 2011
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 ******************************************************************************/
package org.n52.security.common.crypto;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.math.BigInteger;
import java.security.AlgorithmParameters;
import java.security.GeneralSecurityException;
import java.security.Key;
import java.security.KeyFactory;
import java.security.PrivateKey;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.spec.KeySpec;
import java.security.spec.PKCS8EncodedKeySpec;
import java.util.HashMap;
import java.util.Map;

import javax.crypto.Cipher;
import javax.crypto.EncryptedPrivateKeyInfo;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.PBEKeySpec;

import org.apache.commons.codec.binary.Base64InputStream;
import org.n52.security.common.util.FileFinder;

/**
 * Looks for private keys and their according certificates in a file system folder.
 * <p/>
 * Certificates need to be base64 + DER-encoded files with the <code>.cer</code> file name extension.
 * <p/>
 * Private keys need to be base64 + PKCS #8-encoded with the <code>.key</code> file name extension.
 * <p/>
 * A certificate and its according private key have to share a common file name that is used as the alias name during
 * key pair lookup.
 *
 *
 * @author jdr
 *
 */
public class FilesystemKeyPairProvider implements KeyPairProvider {

    private String m_filePath = "";

    private Map<String, String> m_passphrasesByAlias = new HashMap<String, String>();

    private File m_keyPairsDirectory;

    public void init() {
        m_keyPairsDirectory = new FileFinder(m_filePath).getFile();
    }

    @Override
    public KeyPair resolveByAlias(final String alias, char[] password) throws KeyPairResolvingException {

        char[] checkedPassword = lookupPasswordIfNotSpecified(alias, password);

        File certFile = new File(m_keyPairsDirectory, alias + ".cer");

        assertFileExists(certFile);

        PrivateKey privateKey;
        Certificate cert;
        try {
            cert = readCert(certFile);
            String algorithm = cert.getPublicKey().getAlgorithm();
            privateKey = readPrivateKey(alias, checkedPassword, algorithm);
        } catch (Exception e) {
            throw new KeyPairResolvingException("", e);
        }

        return createKeyPair(alias, privateKey, cert);

    }

    private char[] lookupPasswordIfNotSpecified(final String alias, char[] password) {

        String passwordString = m_passphrasesByAlias.get(alias);
        if (passwordString != null && password == null) {
            return passwordString.toCharArray();
        }

        return password;
    }

    private DefaultKeyPair createKeyPair(final String alias, PrivateKey privateKey, Certificate cert) {
        DefaultKeyPair keyPair = new DefaultKeyPair();
        keyPair.setAlias(alias);
        keyPair.setPrivateKey(privateKey);
        keyPair.setCertificate(cert);
        return keyPair;
    }

    private PrivateKey readPrivateKey(final String alias, char[] password, String algorithm) throws IOException,
            GeneralSecurityException {

        File privateKeyFile = new File(m_keyPairsDirectory, alias + ".key");
        if (!privateKeyFile.exists())
            return null;

        byte[] decodeBase64 = decodeBase64(privateKeyFile);

        KeySpec keySpec = new PKCS8EncodedKeySpec(decodeBase64);

        if (password != null) {
            keySpec = decode(decodeBase64, password);
        }
        return KeyFactory.getInstance(algorithm).generatePrivate(keySpec);
    }

    /**
     * @param certFile
     * @return
     * @throws CertificateException
     * @throws IOException
     */
    private X509Certificate readCert(File certFile) throws CertificateException, IOException {
        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        Base64InputStream base64InputStream = new Base64InputStream(new FileInputStream(certFile));
        return (X509Certificate) cf.generateCertificate(base64InputStream);
    }

    private KeySpec decode(byte[] decodeBase64, char[] password) throws GeneralSecurityException, IOException {
        EncryptedPrivateKeyInfo encryptedPrivateKeyInfo = new EncryptedPrivateKeyInfo(decodeBase64);
        Cipher cipher = Cipher.getInstance(encryptedPrivateKeyInfo.getAlgName());
        PBEKeySpec pbeKeySpec = new PBEKeySpec(password);
        SecretKeyFactory secFac = SecretKeyFactory.getInstance(encryptedPrivateKeyInfo.getAlgName());
        Key pbeKey = secFac.generateSecret(pbeKeySpec);
        AlgorithmParameters algParams = encryptedPrivateKeyInfo.getAlgParameters();
        cipher.init(Cipher.DECRYPT_MODE, pbeKey, algParams);
        return encryptedPrivateKeyInfo.getKeySpec(cipher);
    }

    private byte[] decodeBase64(File privateKeyFile) throws IOException {
        Base64InputStream base64InputStream;
        base64InputStream = new Base64InputStream(new FileInputStream(privateKeyFile));

        ByteArrayOutputStream decodedBytes = new ByteArrayOutputStream((int) privateKeyFile.length() * 2);

        byte[] readBuffer = new byte[1024];
        int bytesRead;
        while ((bytesRead = base64InputStream.read(readBuffer)) != -1) {
            decodedBytes.write(readBuffer, 0, bytesRead);
        }
        return decodedBytes.toByteArray();
    }

    private void assertFileExists(File fileToCheck) throws IllegalArgumentException {
        if (!fileToCheck.exists())
            throw new IllegalArgumentException("There is no file <" + fileToCheck.getAbsolutePath() + ">");
    }

    @Override
    public KeyPair resolveBySerialIssuer(String issuerDN, BigInteger serial, char[] password) {

        File certificateFile = findMatchingCertificateFile(issuerDN, serial);

        if (certificateFile == null) {
            return null;
        }
        String alias = certificateFile.getName().replaceAll("\\.cer$", "");

        return resolveByAlias(alias, password);

    }

    private File[] getAllCertificateFiles() {
        File[] certFiles = m_keyPairsDirectory.listFiles(new FilenameFilter() {

            @Override
            public boolean accept(File dir, String name) {
                return name.endsWith(".cer");
            }
        });
        return certFiles;
    }

    private File findMatchingCertificateFile(String issuerDN, BigInteger serial) {
        for (File certFile : getAllCertificateFiles()) {
            X509Certificate readCert;
            try {
                readCert = readCert(certFile);
            } catch (Exception e) {
                throw new KeyPairResolvingException("", e);
            }

            X509Name certIssuerName = new X509Name(readCert.getIssuerDN().getName());
            X509Name compareIssuerName = new X509Name(issuerDN);

            if (certIssuerName.equals(compareIssuerName) && readCert.getSerialNumber().equals(serial)) {
                return certFile;
            }
        }
        return null;
    }

    @Override
    public KeyPair resolveByCertificate(Certificate cert, char[] password) throws KeyPairResolvingException {

        if (cert instanceof X509Certificate) {
            X509Certificate x509Cert = (X509Certificate) cert;
            return resolveBySerialIssuer(x509Cert.getIssuerDN().getName(), x509Cert.getSerialNumber(), password);
        } else {
            return null;
        }
    }

    public String getFilePath() {
        return m_filePath;
    }

    public void setFilePath(String filePath) {
        m_filePath = filePath;
    }

    public void setPassphrasesByAlias(Map<String, String> passphrasesByAlias) {
        m_passphrasesByAlias = passphrasesByAlias;
    }

    public Map<String, String> getPassphrasesByAlias() {
        return m_passphrasesByAlias;
    }

}
