package org.n52.security.common.subject;

import java.util.Set;

import javax.security.auth.Subject;

/**
 * Helper for retrieving credentials from a subject.
 */
public class SubjectCredentialAnalyzer {
    private Subject m_subject;

    public SubjectCredentialAnalyzer(Subject subject) {
        super();
        m_subject = subject;
    }

    public Subject getSubject() {
        return m_subject;
    }

    public <T> T getPublicCredential(Class<T> credentialClass) {
        return getFirstItem(m_subject.getPublicCredentials(credentialClass));
    }

    public <T> T getPrivateCredential(Class<T> credentialClass) {
        return getFirstItem(m_subject.getPrivateCredentials(credentialClass));
    }

    public <T> T getCredential(Class<T> credentialClass) {
        T credential = getPublicCredential(credentialClass);
        if (credential != null) {
            return credential;
        }
        return getPrivateCredential(credentialClass);
    }

    public SubjectCredentialAnalyzer replacePrivateCredential(Object credential) {
        removePrivateCredentials(credential.getClass());
        m_subject.getPrivateCredentials().add(credential);
        return this;
    }

    public SubjectCredentialAnalyzer replacePublicCredential(Object credential) {
        removePublicCredentials(credential.getClass());
        m_subject.getPublicCredentials().add(credential);
        return this;
    }

    public <T> SubjectCredentialAnalyzer removeCredentials(Class<T> credentialClass) {
        removePublicCredentials(credentialClass);
        removePrivateCredentials(credentialClass);
        return this;
    }

    public <T> SubjectCredentialAnalyzer removePublicCredentials(Class<T> credentialClass) {
        Set<Object> credentials = m_subject.getPublicCredentials();
        Set<T> oldCredentials = m_subject.getPublicCredentials(credentialClass);
        removeFromSet(credentials, oldCredentials);
        return this;
    }

    public <T> SubjectCredentialAnalyzer removePrivateCredentials(Class<T> credentialClass) {
        Set<Object> credentials = m_subject.getPrivateCredentials();
        Set<T> oldCredentials = m_subject.getPrivateCredentials(credentialClass);
        removeFromSet(credentials, oldCredentials);
        return this;
    }

    private <T> void removeFromSet(Set<Object> set, Set<T> credentialsToRemove) {
        for (T old : credentialsToRemove) {
            set.remove(old);
        }
    }

    private <T> T getFirstItem(Set<T> setOfItems) {
        return setOfItems.isEmpty() ? null : setOfItems.iterator().next();
    }

}
