/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 22.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.InputSource;

/**
 * Class ClasspathLSResourceResolver resolves resources based on the entity-mapping file.
 * <p/>
 * In contrast to the quite mode of the EnityResolver, the quite mode of a resource resolver,
 * only ensures, that unknown xsd mappings are never resolved,
 * this means that a xsd validator will never contain warnings about non resloved xsd files.
 * So be carefull with the quite mode of the resource resolver.
 *
 * @author Marko Reiprecht
 *         Created on 23.03.2008
 * @version $Revision: $
 * @see org.n52.security.common.xml.ClasspathEntityResolver
 */
public class ClasspathLSResourceResolver extends ClasspathEntityResolver implements LSResourceResolver {

    // ------------------------------ FIELDS ------------------------------

    /**
     * The Logger.
     */
    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(ClasspathLSResourceResolver.class);

    // --------------------------- CONSTRUCTORS ---------------------------

    /**
     * @see ClasspathEntityResolver#ClasspathEntityResolver()
     */
    public ClasspathLSResourceResolver() {
    }

    /**
     * Constructor creates a new instance.
     *
     * @param quite the quite mode.
     */
    public ClasspathLSResourceResolver(final boolean quite) {
        super(quite);
    }

    /**
     * Constructor creates a new instance.
     *
     * @param quite  the quite mode.
     * @param loader the classloader to use, if null this.getClass().getClassLoader() is used.
     */
    public ClasspathLSResourceResolver(final boolean quite, final ClassLoader loader) {
        super(quite, loader);
    }

    // ------------------------ INTERFACE METHODS ------------------------


    // --------------------- Interface LSResourceResolver ---------------------

    /**
     * Method resolveResource resolves the resource by looking for special mapping file.
     * <p/>
     * First it looks for a namespaceURI mapping, then for a publicId mapping and then for a systemId mapping.
     * If none mapping is found it returns a "emtpy" inputsource in quite mode or null in nonquite mode.
     *
     * @param type         the type.
     * @param namespaceURI the namespaceURI.
     * @param publicId     the publicId.
     * @param systemId     the systemId.
     * @param baseURI      the baseURI.
     * @return LSInput.
     */
    public LSInput resolveResource(final String type,
                                   final String namespaceURI,
                                   final String publicId,
                                   final String systemId,
                                   final String baseURI) {
        if (LOG.isTraceEnabled()) {
            LOG.trace(
                    "ClasspathLSResourceResolver.resolveResource("
                    + type + ","
                    + namespaceURI + ","
                    + publicId + ","
                    + systemId + ","
                    + baseURI + ")");
        }
        String resourcefile = null;

        if (resourcefile == null && systemId != null) {
            resourcefile = getMappedResourceFileName(systemId);
        }
        if (resourcefile == null && publicId != null) {
            resourcefile = getMappedResourceFileName(publicId);
        }
        if (resourcefile == null && namespaceURI != null) {
            resourcefile = getMappedResourceFileName(namespaceURI);
        }

        if (resourcefile != null) {
            InputSource resolvedsource = createInputSource(resourcefile);
            if (resolvedsource != null) {
                return toLSInput(resolvedsource,baseURI!=null? baseURI : systemId);
            }
        }
        // in quite mode simple return a empty inputsource,
        // in contrast to the emptyInputSource() method for entities,
        // here a real "empty" inputsource must be returned,
        // because a valid xml schema (xml file) is expected by the parser.
        return isQuite() ? toLSInput(new InputSource(),null) : null;
    }

    // -------------------------- OTHER METHODS --------------------------

    /**
     * Method toLSInput converts a InputSource to a LSInput.
     *
     * @param inputSource the inputSource.
     * @return LSInput.
     */
    private LSInput toLSInput(final InputSource inputSource,String baseUri) {
        return new LSInputSourceAdapter(inputSource, baseUri);
    }
}
