/*******************************************************************************
 * Copyright (C) 2008
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 *
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 *
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 *
 * Author: Marko Reiprecht
 * Created: 24.03.2008
 ******************************************************************************/
/* $Id: $ */
package org.n52.security.common.xml;

import javax.xml.validation.Schema;

/**
 * Class DOMParserOptions are options describing some common options for DOMParsers.
 * 
 * @author Marko Reiprecht
 *         Created on 24.03.2008
 * @version $Revision: $
 */
public class DOMParserOptions {

    private boolean m_namespaceaware = true;
    private boolean m_validating = false;
    private boolean m_ignoreComments = true;
    private boolean m_ignoreElementContentWhitespace = false;
    private boolean m_xincludeAware = false;
    private boolean m_coalescing = false;
    private boolean m_expandEntityReferences = false;
    private Schema m_schema = null;

    /**
     * Method gets the default options.
     * 
     * The default is a namespaceaware, non validating parser, which ignores comments.
     * 
     * @return DOMParserOptions.
     */
    public static DOMParserOptions getDefault() {
        return new DOMParserOptions();
    }

    /**
     * Method sets the ignoreComments to false.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions notIgnoreComments() {
        setIgnoreComments(false);
        return this;
    }

    /**
     * Method sets the ignoreComments flag.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions ignoreComments() {
        setIgnoreComments(true);
        return this;
    }

    /**
     * Method gets it comments are ignored.
     * 
     * @return boolean.
     */
    public boolean isIgnoreComments() {
        return m_ignoreComments;
    }

    /**
     * Method sets the ignore comments flag.
     * 
     * @param ignorecomments
     *            the ignoreComments.
     */
    public void setIgnoreComments(final boolean ignorecomments) {
        m_ignoreComments = ignorecomments;
    }

    /**
     * Method sets the namespaceaware flag.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions namespaceaware() {
        setNamespaceaware(true);
        return this;
    }

    /**
     * Method sets the namespaceaware flag to false.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions notNamespaceaware() {
        setNamespaceaware(false);
        return this;
    }

    /**
     * Method gets the parser is namespaceaware.
     * 
     * @return boolean.
     */
    public boolean isNamespaceaware() {
        return m_namespaceaware;
    }

    /**
     * Method sets the namespaceaware flag.
     * 
     * @param namespaceaware
     *            the namespaceaware.
     */
    public void setNamespaceaware(final boolean namespaceaware) {
        m_namespaceaware = namespaceaware;
    }

    /**
     * Method sets the validating to false.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions notValidating() {
        setValidating(false);
        return this;
    }

    /**
     * Method sets the validating flag.
     * 
     * @return DOMParserOptions.
     */
    public DOMParserOptions validating() {
        setValidating(true);
        return this;
    }

    /**
     * Method gets if the parser is in validating mode.
     * Consider that validating works only in xml files based on dtds.
     * To validate xml files based on xml schemata use the schema class.
     * 
     * @return boolean.
     */
    public boolean isValidating() {
        return m_validating;
    }

    /**
     * Method sets the validating flag.
     * 
     * @param validating
     *            the validating.
     */
    public void setValidating(final boolean validating) {
        m_validating = validating;
    }

    public DOMParserOptions ignoreElementContentWhitespace() {
        setIgnoreElementContentWhitespace(true);
        return this;
    }

    public DOMParserOptions notIgnoreElementContentWhitespace() {
        setIgnoreElementContentWhitespace(false);
        return this;
    }

    public boolean isIgnoreElementContentWhitespace() {
        return m_ignoreElementContentWhitespace;
    }

    public void setIgnoreElementContentWhitespace(boolean ignoreElementWhitespace) {
        m_ignoreElementContentWhitespace = ignoreElementWhitespace;
    }

    public DOMParserOptions xincludeAware() {
        setXincludeAware(true);
        return this;
    }

    public DOMParserOptions notXincludeAware() {
        setXincludeAware(false);
        return this;
    }

    public boolean isXincludeAware() {
        return m_xincludeAware;
    }

    public void setXincludeAware(boolean xincludeAware) {
        m_xincludeAware = xincludeAware;
    }

    public DOMParserOptions coalescing() {
        setCoalescing(true);
        return this;
    }

    public DOMParserOptions notCoalescing() {
        setCoalescing(false);
        return this;
    }

    public boolean isCoalescing() {
        return m_coalescing;
    }

    public void setCoalescing(boolean coalescing) {
        m_coalescing = coalescing;
    }

    public DOMParserOptions expandEntityReferences() {
        setExpandEntityReferences(true);
        return this;
    }

    public DOMParserOptions notExpandEntityReferences() {
        setExpandEntityReferences(false);
        return this;
    }

    public boolean isExpandEntityReferences() {
        return m_expandEntityReferences;
    }

    public void setExpandEntityReferences(boolean expandEntityReferences) {
        m_expandEntityReferences = expandEntityReferences;
    }

    public DOMParserOptions schema(Schema schema) {
        setSchema(schema);
        return this;
    }

    public Schema getSchema() {
        return m_schema;
    }

    public void setSchema(Schema schema) {
        m_schema = schema;
    }

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + (m_coalescing ? 1231 : 1237);
        result = prime * result + (m_expandEntityReferences ? 1231 : 1237);
        result = prime * result + (m_ignoreComments ? 1231 : 1237);
        result = prime * result + (m_ignoreElementContentWhitespace ? 1231 : 1237);
        result = prime * result + (m_namespaceaware ? 1231 : 1237);
        result = prime * result + ((m_schema == null) ? 0 : m_schema.hashCode());
        result = prime * result + (m_validating ? 1231 : 1237);
        result = prime * result + (m_xincludeAware ? 1231 : 1237);
        return result;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        DOMParserOptions other = (DOMParserOptions) obj;
        if (m_coalescing != other.m_coalescing)
            return false;
        if (m_expandEntityReferences != other.m_expandEntityReferences)
            return false;
        if (m_ignoreComments != other.m_ignoreComments)
            return false;
        if (m_ignoreElementContentWhitespace != other.m_ignoreElementContentWhitespace)
            return false;
        if (m_namespaceaware != other.m_namespaceaware)
            return false;
        if (m_schema == null) {
            if (other.m_schema != null)
                return false;
        } else if (!m_schema.equals(other.m_schema))
            return false;
        if (m_validating != other.m_validating)
            return false;
        if (m_xincludeAware != other.m_xincludeAware)
            return false;
        return true;
    }

    @Override
    public String toString() {
        return String
                .format("DOMParserOptions [namespaceaware=%s, validating=%s, ignoreComments=%s, ignoreElementContentWhitespace=%s, xincludeAware=%s, coalescing=%s, expandEntityReferences=%s, schema=%s]",
                        m_namespaceaware, m_validating, m_ignoreComments, m_ignoreElementContentWhitespace,
                        m_xincludeAware, m_coalescing, m_expandEntityReferences, m_schema);
    }

}
