package org.n52.security.common.xml;

import static org.n52.security.common.collections.CollectionUtil.addAll;
import static org.n52.security.common.collections.CollectionUtil.asList;
import static org.n52.security.common.collections.CollectionUtil.transform;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.XMLConstants;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventFactory;
import javax.xml.stream.XMLEventReader;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.Namespace;
import javax.xml.stream.events.StartElement;
import javax.xml.stream.events.XMLEvent;

import org.n52.security.common.collections.AbstractTransformerFilter;
import org.n52.security.common.util.AbstractObjectPool;
import org.xml.sax.InputSource;

/**
 * Iterates over chunks of xml.
 * Because this iterator supports nested chunks, each chunk is returned twice.
 * First a "start" chunk, if the start element of a chunk is seen.
 * Second a "end" chunk, if the end element of a chunk is seen, only here the content is given.
 * 
 * If a chunk contains nested chunks, then the content of the nested chunks is not embedded in the "parent" chunk.
 * 
 * When an element is a start/end of a chunk is open to {@link Matcher} implementations. The default way is to specify
 * element QNames.
 * 
 * Please note that the iterator updates the content of the Parent-Chunks if a nested chunk is visited.
 * Simple use the {@link XMLChunk#getParent()} method to get an updated parent.
 */
public class XMLChunkIterator implements Iterator<XMLChunk> {

    /**
     * a Matcher decides if an element at a given "xpath" is the start of chunk.
     */
    private static interface Matcher {
        boolean matches(List<QName> elementPath);
    }

    /**
     * Matches an element path, if the last element in the path matches a given qname.
     */
    public static class QNameMatcher implements Matcher {
        private QName m_qname;

        public QNameMatcher(QName qname) {
            m_qname = qname;
        }

        @Override
        public boolean matches(List<QName> elementPath) {
            return elementPath.get(elementPath.size() - 1).equals(m_qname);
        }
    }

    /**
     * Contains a list of matchers, which say if an element is a start of a chunk.
     */
    private final List<Matcher> m_elementMatcher = new ArrayList<Matcher>();

    private XMLEventReader m_reader;
    private XMLChunk m_currentChunk;
    private XMLEventFactory m_xmlEventFactory;
    private ChunkWriterPool m_chunkWriterPool;
    private LinkedList<Chunk> m_chunkStack;
    private LinkedList<ElementNamespaces> m_namespaceStack;

    private LinkedList<QName> m_elementXPath;

    public XMLChunkIterator(InputSource source, Iterable<Matcher> elementMatcher) {
        addAll(m_elementMatcher, elementMatcher);
        // create event reader from the input source.
        try {
            m_reader = createEventReader(source);
        } catch (XMLStreamException e) {
            throw new IllegalStateException("Couldn't open the xml stream!", e);
        }
        init();
        m_currentChunk = readNext();
    }

    public XMLChunkIterator(InputSource source, Matcher... elementMatcher) {
        this(source, asList(elementMatcher));
    }

    public XMLChunkIterator(InputSource source, QName... chunkStartElements) {
        this(source, transform(asList(chunkStartElements), new AbstractTransformerFilter<Matcher, QName>() {
            @Override
            public Matcher transform(QName qname) {
                return new QNameMatcher(qname);
            }
        }));
    }

    @Override
    public boolean hasNext() {
        return m_currentChunk != null;
    }
    @Override
    public XMLChunk next() {
        XMLChunk chunk = m_currentChunk;
        m_currentChunk = readNext();
        return chunk;
    }

    @Override
    public void remove() {
        throw new UnsupportedOperationException("<remove> is not implemented on a XMLChunkIterator");
    }

    private XMLChunk readNext() {
        try {
            while (m_reader.hasNext()) {
                XMLEvent evt = m_reader.nextEvent();
                XMLChunk chunk = null;

                if (evt.isStartElement()) {
                    StartElement startElem = (StartElement) evt;
                    List<QName> path = pushElementXPath(startElem);
                    boolean isChunkElement = isChunkStart(path);
                    evt = pushNamespaces(startElem, isChunkElement);
                    if (isChunkElement) {
                        chunk = startChunk(path);
                    }
                }
                writeToChunk(evt);
                if (evt.isEndElement()) {
                    if (isChunkStart(m_elementXPath)) {
                        chunk = endChunk();
                    }
                    popNamespaces();
                    popElementXPath();
                }
                if (chunk != null) {
                    return chunk;
                }
            }
            // really reach the end
            release();
            return null;
        } catch (Exception e) {
            try {
                release();
            } catch (Exception ex) {
                // ignore
            }
            throw new DOMParserException("error reading xml chunk", e);
        }
    }

    private void init() {
        // create event factory to create new xml events
        m_xmlEventFactory = XMLEventFactory.newInstance();
        // create a dynamicaly grown writer pool
        m_chunkWriterPool = new ChunkWriterPool();
        // create handler stack
        m_chunkStack = new LinkedList<Chunk>();
        // create namespace stack
        m_namespaceStack = new LinkedList<ElementNamespaces>();
        // create xml path stack
        m_elementXPath = new LinkedList<QName>();
    }

    private void release() {
        try {
            m_reader.close();
        } catch (XMLStreamException e) {
            // ignore
        }
        m_chunkStack.clear();
        m_elementMatcher.clear();
        m_namespaceStack.clear();
        m_reader = null;
        m_xmlEventFactory = null;
    }

    private void writeToChunk(XMLEvent evt) throws XMLStreamException {
        if (!m_chunkStack.isEmpty()) {
            m_chunkStack.getFirst().getWriter().write(evt);
        }
    }

    private XMLChunk startChunk(List<QName> path) throws XMLStreamException {
        Chunk parentChunk = m_chunkStack.isEmpty() ? null : m_chunkStack.getFirst();
        XMLChunk parent = null;
        if (parentChunk != null) {
            XMLChunk old = parentChunk.getXmlChunk();
            parent = new XMLChunk(true, old.getXpath(), parentChunk.getWriter().getChunkContent(), old
                    .getParent());
            // replace old parent chunk with new content which is maybe before the child element
            parentChunk.setXmlChunk(parent);
        }

        Chunk chunk = new Chunk(new XMLChunk(true, asList(path), parent), m_chunkWriterPool.acquire());
        m_chunkStack.addFirst(chunk);
        return chunk.getXmlChunk();
    }
    private XMLChunk endChunk() throws XMLStreamException {
        Chunk chunk = m_chunkStack.removeFirst();
        String content = chunk.getWriter().getChunkContent();
        m_chunkWriterPool.release(chunk.getWriter());

        Chunk parentChunk = m_chunkStack.isEmpty() ? null : m_chunkStack.getFirst();

        return new XMLChunk(false, chunk.getXmlChunk().getXpath(), content,
                parentChunk != null ? parentChunk.getXmlChunk() : null);
    }

    private boolean isChunkStart(List<QName> path) {
        for (Matcher m : m_elementMatcher) {
            if (m.matches(path)) {
                return true;
            }
        }
        return false;
    }

    private List<QName> pushElementXPath(StartElement elem) {
        QName qName = elem.getName();
        m_elementXPath.addLast(qName);
        return m_elementXPath;
    }

    private void popElementXPath() {
        m_elementXPath.removeLast();
    }

    private XMLEvent pushNamespaces(final StartElement elem, boolean isChunkStart) throws XMLStreamException {
        @SuppressWarnings("unchecked")
        ElementNamespaces namespaces = new ElementNamespaces(
                elem.getNamespaces(),
                m_namespaceStack.isEmpty()
                        ? Collections.<Namespace>emptySet().iterator()
                        : m_namespaceStack.getFirst().getAllAvailableNamespaces(),
                m_xmlEventFactory);
        m_namespaceStack.addFirst(namespaces);
        // we have to append the start element with all available namespaces
        // otherwise the element is not correct
        // and XMLOutputFactory.IS_REPAIRING_NAMESPACES, does not work,
        // because some elements are skiped
        // it also optimizes the namespace declarations, no namespace declarations are repeated!
        return m_xmlEventFactory.createStartElement(elem.getName(), elem.getAttributes(),
                isChunkStart ? namespaces.getAllAvailableNamespaces() : namespaces.getLocalNamespaces());
    }

    private void popNamespaces() {
        m_namespaceStack.removeFirst();
    }

    /**
     * Creates the XMLEventReader working on the given inputsource.
     * 
     */
    private XMLEventReader createEventReader(final InputSource source)
            throws XMLStreamException {
        // Create the input reader to read the source.
        XMLInputFactory inFactory = XMLInputFactory.newInstance();
        inFactory.setProperty(XMLInputFactory.IS_NAMESPACE_AWARE, Boolean.TRUE);
        // in jdk 1.6 this method is throws an Malformed URL exception, so we must use
        // the plain methods
        // XMLEventReader reader = inFactory.createXMLEventReader(new SAXSource(source));
        final XMLEventReader reader;
        if (source.getByteStream() != null) {
            if (source.getEncoding() != null && source.getEncoding().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getByteStream(), source.getEncoding());
            } else if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getByteStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getByteStream());
            }
        } else if (source.getCharacterStream() != null) {
            if (source.getSystemId() != null && source.getSystemId().length() > 0) {
                reader = inFactory.createXMLEventReader(source.getSystemId(), source.getCharacterStream());
            } else {
                reader = inFactory.createXMLEventReader(source.getCharacterStream());
            }
        } else {
            throw new IllegalArgumentException("InputSource holds no valid input");
        }
        return reader;
    }

    private static final class ElementNamespaces {

        private Map<String, Namespace> m_localNamespaces = new HashMap<String, Namespace>();
        private Map<String, Namespace> m_allAvailableNamespaces = new HashMap<String, Namespace>();

        private ElementNamespaces(
                final Iterator<Namespace> elementNamespaces,
                final Iterator<Namespace> parentNamespaces,
                final XMLEventFactory fac) {
            parseNamespaces(elementNamespaces, parentNamespaces, fac);
        }

        private void parseNamespaces(
                final Iterator<Namespace> elementNamespaces,
                final Iterator<Namespace> parentNamespaces,
                final XMLEventFactory fac) {
            addToLocalNamespaces(elementNamespaces);
            while (parentNamespaces.hasNext()) {
                Namespace ns = parentNamespaces.next();
                m_allAvailableNamespaces.put(ns.getPrefix(), ns);
            }
            // this fixes a bug in the codehaus stax implementation, where xml:lang attributes not correctly written
            // if the xmlns:xml=".." is not emited!
            if (!m_allAvailableNamespaces.containsKey(XMLConstants.XML_NS_PREFIX)) {
                m_allAvailableNamespaces.put(XMLConstants.XML_NS_PREFIX,
                        fac.createNamespace(XMLConstants.XML_NS_PREFIX, XMLConstants.XML_NS_URI));
            }

            // remove duplicated local namespaces
            removeAllreadyAvailableNamespacesFromLocal();
            m_allAvailableNamespaces.putAll(m_localNamespaces);
        }

        private void removeAllreadyAvailableNamespacesFromLocal() {
            for (Iterator<String> iterator = m_localNamespaces.keySet().iterator(); iterator.hasNext();) {
                String prefix = iterator.next();
                // test for duplicated prefix + uri
                // be carefull if you check namespace.equals(namespace) and the provider (till jdk6_12) has wrong
                // implemented the equals method than the test fails
                // so here is excplicite checked for prefix and namespace
                if (m_allAvailableNamespaces.containsKey(prefix)
                        && m_allAvailableNamespaces.get(prefix).getNamespaceURI().equals(
                                m_localNamespaces.get(prefix).getNamespaceURI())) {
                    iterator.remove();
                }
            }
        }

        private void addToLocalNamespaces(final Iterator<Namespace> namespaces) {
            while (namespaces.hasNext()) {
                Namespace namespace = namespaces.next();
                m_localNamespaces.put(namespace.getPrefix(), namespace);
            }
        }

        public Iterator<Namespace> getLocalNamespaces() {
            return m_localNamespaces.values().iterator();
        }

        public Iterator<Namespace> getAllAvailableNamespaces() {
            return m_allAvailableNamespaces.values().iterator();
        }
    }

    private static final class Chunk {

        private XMLChunk m_xmlChunk;
        private final ChunkWriter m_writer;

        private Chunk(final XMLChunk xmlChunk, final ChunkWriter writer) throws XMLStreamException {
            m_xmlChunk = xmlChunk;
            m_writer = writer;
        }

        public XMLChunk getXmlChunk() {
            return m_xmlChunk;
        }

        public void setXmlChunk(XMLChunk xmlChunk) {
            m_xmlChunk = xmlChunk;
        }

        public ChunkWriter getWriter() {
            return m_writer;
        }

    }

    private static final class ChunkWriterPool extends AbstractObjectPool<ChunkWriter, Void> {

        public ChunkWriterPool() {
            super(3);
        }

        public ChunkWriter acquire() {
            return this.acquire(null);
        }
        public void release(ChunkWriter w) {
            w.release();
            this.release(w, null);
        }

        @Override
        public ChunkWriter create(Void key) {
            try {
                return new ChunkWriter();
            } catch (XMLStreamException e) {
                throw new IllegalStateException("can't create xml chunk writer!", e);
            }
        }

    }

    private static final class ChunkWriter {

        private ResettableStringWriter m_stringWriter;
        private XMLEventWriter m_eventWriter;
        private XMLOutputFactory m_eventWriterFactory;

        public ChunkWriter() throws XMLStreamException {
            m_stringWriter = new ResettableStringWriter();
            m_eventWriterFactory = XMLOutputFactory.newInstance();
        }

        public void write(XMLEvent evt) throws XMLStreamException {
            getEventWriter().add(evt);
        }

        public String getChunkContent() throws XMLStreamException {
            if (m_eventWriter != null) {
                m_eventWriter.flush();
            }
            return m_stringWriter.getString();
        }

        private XMLEventWriter getEventWriter() throws XMLStreamException {
            if (m_eventWriter == null) {
                m_stringWriter.reset();
                m_eventWriter = m_eventWriterFactory.createXMLEventWriter(m_stringWriter);
            }
            return m_eventWriter;
        }

        public void release() {
            try {
                m_eventWriter.close();
            } catch (XMLStreamException e) {
                // ignore
            }
            m_eventWriter = null;
            m_stringWriter.reset();
        }
    }

    /**
     * A Writer that buffers the data in memory as a string. The buffer can be reset at any time and the contents of the
     * buffer will be returned.
     */
    private static final class ResettableStringWriter extends Writer {

        /**
         * The internal buffer used to store the data written to this stream.
         */
        private StringBuilder m_buffer = new StringBuilder(1024);

        /**
         * The maximum number of chars that this writer will reuse. (2MB)
         */
        private int m_maxChars = 1024 * 1024;

        /**
         * Constructor creates a new instance.
         */
        private ResettableStringWriter() {}

        /**
         * Constructor creates a new instance.
         * 
         * @param maxChars
         *            the maxChars this buffer will reuse.
         */
        private ResettableStringWriter(final int maxChars) {
            m_maxChars = maxChars;
        }

        public void close() throws IOException {
            // no op
        }

        public void flush() throws IOException {
            // no op
        }

        public void write(final char[] cbuf, final int off, final int len) throws IOException {
            m_buffer.append(cbuf, off, len);
        }

        /**
         * Method gets the buffer.
         * 
         * @return StringBuffer.
         */
        public String getString() {
            return m_buffer.substring(0);
        }

        /**
         * Resets the stream to a length of zero. Any data written to the stream since the last call to reset will be
         * returned.
         * 
         * @return the contents of the stream buffer
         */
        public ResettableStringWriter reset() {
            if (m_buffer.length() > m_maxChars) {
                m_buffer = new StringBuilder(1024);
            }
            m_buffer.setLength(0);
            return this;
        }
    }
}
