/*******************************************************************************
 * Copyright (C) 2009
 * by 52 North Initiative for Geospatial Open Source Software GmbH
 * 
 * Contact: Andreas Wytzisk
 * 52 North Initiative for Geospatial Open Source Software GmbH
 * Martin-Luther-King-Weg 24
 * 48155 Muenster, Germany
 * info@52north.org
 * 
 * This program is free software; you can redistribute and/or modify it under
 * the terms of the GNU General Public License version 2 as published by the
 * Free Software Foundation.
 * 
 * This program is distributed WITHOUT ANY WARRANTY; even without the implied
 * WARRANTY OF MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program (see gnu-gpl v2.txt). If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA or
 * visit the Free Software Foundation web page, http://www.fsf.org.
 ******************************************************************************/

package org.n52.security.common.xml;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.XMLConstants;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

/**
 * This utility removes all xmlns attribute declarations from the dom, but preserves indirect referenced namespaces e.g.
 * linked via attribute/element content( xsi:type="prefix:...").
 * 
 * @author Marko Reiprecht
 */
public class XMLNamespaceAttributeDeclarationRemover {
    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory
            .getLogger(XMLNamespaceAttributeDeclarationRemover.class);
    private static final Pattern NSPREFIX_PATTERN = Pattern.compile("^([a-zA-Z0-1_\\-\\.]+):[a-zA-Z0-1_\\-\\.]+$");

    /**
     * Method removes the namespace declarations.
     * 
     * @param document
     *            the document.
     * @return Document.
     */
    public Document removeNSDecls(final Document document) {
        removeNSDecls(document.getDocumentElement());
        return document;
    }

    /**
     * Method removes all namespace attributes.
     * 
     * @param node
     *            the root.
     * @return Element.
     */
    public Element removeNSDecls(final Element node) {
        new DocumentTraverser().traverseDepthFirst(node, new DocumentVisitorAdapter() {
            private Matcher m_matcher = NSPREFIX_PATTERN.matcher("");
            /**
             * collect all found namespaces.
             */
            private Map<String, Attr> m_foundNamespaces = new HashMap<String, Attr>();
            /**
             * lists required namespaces (used in attribute values).
             */
            private Set<String> m_requiredNamespacePrefixes = new HashSet<String>();

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#visit(org.w3c.dom.Attr)
             */
            public void visit(final Attr attr) {
                String name = attr.getName();
                String value = attr.getValue();
                if (name.startsWith(XMLConstants.XMLNS_ATTRIBUTE)) {
                    m_foundNamespaces.put(name.replace(XMLConstants.XMLNS_ATTRIBUTE + ":", ""), attr);
                    attr.getOwnerElement().removeAttributeNode(attr);
                } else if (value != null && m_matcher.reset(value).matches()) {
                    m_requiredNamespacePrefixes.add(m_matcher.group(1));
                }
            }

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#postVisit(org.w3c.dom.Text)
             */
            public void visit(Text text) {
                String value = text.getNodeValue();
                if (value != null && m_matcher.reset(value).matches()) {
                    m_requiredNamespacePrefixes.add(m_matcher.group(1));
                }
            }

            /**
             * @see org.n52.security.common.xml.DocumentVisitorAdapter#postVisit(org.w3c.dom.Element)
             */
            public void postVisit(Element element) {
                if (element == node) {
                    // end of iteration
                    if (m_requiredNamespacePrefixes.size() > 0) {
                        for (Iterator prefixes = m_requiredNamespacePrefixes.iterator(); prefixes.hasNext();) {
                            String prefix = (String) prefixes.next();
                            Attr decl = (Attr) m_foundNamespaces.get(prefix);
                            if (decl == null) {
                                // look up in parent
                                decl = findXMLNSDeclInParentNode(element, prefix);
                            }
                            if (decl != null) {
                                // append xmlns declaration to root element
                                element.setAttributeNodeNS(decl);
                            } else {
                                if (LOG.isWarnEnabled()) {
                                    LOG.warn("no namespace attribute found in sub tree for prefix <" + prefix + ">");
                                }
                            }
                        }
                    }
                }
            }

            private Attr findXMLNSDeclInParentNode(final Element element, final String prefix) {
                String attrName = XMLConstants.XMLNS_ATTRIBUTE + ":" + prefix;
                Element cur = element;
                while (cur != null) {
                    Attr attr = cur.getAttributeNode(attrName);
                    if (attr != null) {
                        // we don't remove this node, so we have to clone it
                        return (Attr) attr.cloneNode(false);
                    }
                    Node parent = cur.getParentNode();
                    if (Node.ELEMENT_NODE == parent.getNodeType()) {
                        cur = (Element) parent;
                    } else {
                        cur = null;
                    }
                }
                return null;
            }

        });
        return node;
    }
}
